/* Tokenizer
 * Copyright 2009 Jetro Lauha
 *
 * Utility class to tokenize strings.
 * Inspired by "tokenwad" by Jari Komppa.
 *
 * My home page / blog URL:
 *   http://jet.ro
 *
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
 * $Id: Tokenizer.cpp 134 2010-01-24 21:53:42Z tonic $
 * $Revision: 134 $
 */

#include <cstdlib>
#include <cstring>
#include <cctype>
#include "Tokenizer.h"


#ifdef _WIN32
#define strcasecmp _stricmp
#define strdup _strdup
#endif


Tokenizer::Tokenizer(bool trimWhitespace) :
    mTrim(trimWhitespace),
    mCount(0),
    mCopy(0),
    mTokens(0)
{
}


Tokenizer::~Tokenizer()
{
    clear(mTrim);
}


void Tokenizer::clear(bool trimWhitespace)
{
    mTrim = trimWhitespace;
    mCount = 0;
    if (mCopy)
        free(mCopy);
    mCopy = 0;
    if (mTokens)
        delete[] mTokens;
    mTokens = 0;
}


bool Tokenizer::tokenize(const char *str, const char *separators)
{
    clear(mTrim);

    if (str == 0 || separators == 0)
        return false;

    mCopy = strdup(str);
    if (!mCopy)
        return false;

    return internalTokenize(mCopy, separators);
}


bool Tokenizer::tokenizeInPlace(char *str, const char *separators)
{
    clear(mTrim);

    if (str == 0 || separators == 0)
        return false;

    return internalTokenize(str, separators);
}


bool Tokenizer::internalTokenize(char *str, const char *separators)
{
    const int orgStrLen = (int)strlen(str);
    const int sepCount = (int)strlen(separators);
    if (orgStrLen == 0 || sepCount == 0)
        return false;

    // count tokens & allocate space for pointers to them

    mCount = 1;
    for (int a = 0; a < orgStrLen; ++a)
    {
        for (int b = 0; b < sepCount; ++b)
        {
            if (str[a] == separators[b])
            {
                ++mCount;
                break;
            }
        }
    }

    mTokens = new char *[mCount];
    if (!mTokens)
        return false;

    // store pointers to tokens, converting separators to zero-terminators

    int currentToken = 0;
    mTokens[currentToken] = str;
    for (int a = 0; a < orgStrLen; ++a)
    {
        for (int b = 0; b < sepCount; ++b)
        {
            if (str[a] == separators[b])
            {
                str[a] = 0;
                if (++currentToken < mCount)
                    mTokens[currentToken] = str + a + 1;
            }
        }
    }

    // if enabled, trim all whitespace and control chars from begin/end of tokens

    if (mTrim)
    {
        for (int a = 0; a < mCount; ++a)
        {
            int tokenLength;
            if (a == mCount - 1)
                tokenLength = str + orgStrLen - mTokens[a];
            else
                tokenLength = mTokens[a + 1] - mTokens[a];
            // trim beginning
            int pos = 0;
            while (pos < tokenLength && mTokens[a][pos] <= 32)
                ++pos;
            mTokens[a] += pos;
            tokenLength -= pos;
            // trim ending
            pos = tokenLength - 1;
            while (pos >= 0 && mTokens[a][pos] <= 32)
            {
                mTokens[a][pos] = 0; // overwrite ending whitespace to move back zero terminator
                --pos;
            }
        }
    } // mTrim

    return true;
}


int Tokenizer::getAsInt(int index, int defaultValue)
{
    char *token = get(index);
    if (token == 0)
        return defaultValue;
    // skip whitespace and control chars
    while (*token && *token <= 32)
        ++token;
    if (*token == '+' || *token == '-' || (*token >= '0' && *token <= '9'))
        return atoi(token);
    return defaultValue;
}


bool Tokenizer::equals(int index, const char *str)
{
    if (str == 0)
        return false;
    char *token = get(index);
    if (token == 0)
        return false;
    return strcmp(token, str) == 0;
}


bool Tokenizer::equalsIgnoreCase(int index, const char *str)
{
    if (str == 0)
        return false;
    char *token = get(index);
    if (token == 0)
        return false;
    return strcasecmp(token, str) == 0;
}
